/*
      *****************************************************************************
      *  Copyright (C) 2013 by Reinhard Heuberger                                 *
      *                                                                           *
      *  All rights reserved.                                                     *
      *                                                                           *
      *  Redistribution and use in source and binary forms, with or without       *
      *  modification, are permitted provided that the following conditions       *
      *  are met:                                                                 *
      *                                                                           *
      *  1. Redistributions of source code must retain the above copyright        *
      *     notice, this list of conditions and the following disclaimer.         *
      *  2. Redistributions in binary form must reproduce the above copyright     *
      *     notice, this list of conditions and the following disclaimer in the   *
      *     documentation and/or other materials provided with the distribution.  *
      *  3. Neither the name of the author nor the names of its contributors may  *
      *     be used to endorse or promote products derived from this software     *
      *     without specific prior written permission.                            *
      *                                                                           *
      *****************************************************************************
*/

// This module simulates the idle loop pattern as it was implemented on 
// a console like on a PDP-11/70 when no other programs are active. As
// far as I know, it was the RSX-11M idle loop ( or RSTS/E idle loop ).
// If mode = 1 ( = Power-OK signal from RL02 controller ) the idle-loop
// pattern is running using the 4.1MHz clock from the RL02 controller board.
// Otherwise, a simple single-bit shift pattern is executed using another
// clock signal ( ext_clock ) idicating offline mode.
//
//  by: Reinhard Heuberger , www.PDP11GY.com, JUL, 2011
//
module like_1170_console( enable, rlv_clk, ext_clk, mode, shift );
					 
input enable, rlv_clk, ext_clk, mode;
output [15:0] shift;

reg [31:0] tmp1;
reg [31:0] tmp2;
reg [15:0] shift1;
reg [15:0] shift2;
tri [15:0] shift;
reg [7:0] HB;
reg [7:0] LB;

initial
  begin
   LB <= 8'b00001111;       // Preset LOW-Byte
   HB <= 8'b11110000;       // Preset HIGH-Byte
   shift2 <= 16'h0001;      // Preset offline pattern
  end
//
assign shift = (enable && !mode) ? shift1 : 16'bz;
assign shift = (enable && mode)  ? shift2 : 16'bz;
//
//
always @(posedge rlv_clk)    // Counter/Divider( 24:1)
begin
  if (enable)
    begin
     tmp1 <= tmp1 + 1;
    end
end
//
always @(posedge ext_clk)    // Counter/Divider( 24:1)
begin
  if (enable)
    begin
     tmp2 <= tmp2 + 1;
    end
end
//
//
always @(posedge tmp1[18])   // Shift register #1
begin
 if (enable && !mode)
   begin
     LB = LB << 1;           // Shift LOW-Byte
     HB = HB >> 1;           // Shift HIGH-Byte
     LB[0] <= LB[7];         // Rotate LOW-Byte 
     HB[7] <= HB[0];         // Rotate HIGH-Byte
     shift1[15:8] <= HB;     // Copy LOW-Byte and HIGH-Byte    
     shift1[7:0]  <= LB;     //      to the 16Bit vector.
     // shift[7:0] <= ~(LB); // Also works, HB not necessary
   end
 end
 //
always @(posedge tmp2[21])     // Shift register #2 18=@16.4Mhz
 begin
 if (enable && mode)
   begin
     shift2 = shift2 << 1;      // shift word
     shift2[0] <= shift2[15];   // rotate word
     //shift2[0] <= 1;            // Bit 0 remains "1"
   end 
 end
endmodule
