/*
      *****************************************************************************
      *  Copyright (C) by Reinhard Heuberger                                 *
      *                                                                           *
      *  All rights reserved.                                                     *
      *                                                                           *
      *  Redistribution and use in source and binary forms, with or without       *
      *  modification, are permitted provided that the following conditions       *
      *  are met:                                                                 *
      *                                                                           *
      *  1. Redistributions of source code must retain the above copyright        *
      *     notice, this list of conditions and the following disclaimer.         *
      *  2. Redistributions in binary form must reproduce the above copyright     *
      *     notice, this list of conditions and the following disclaimer in the   *
      *     documentation and/or other materials provided with the distribution.  *
      *  3. Neither the name of the author nor the names of its contributors may  *
      *     be used to endorse or promote products derived from this software     *
      *     without specific prior written permission.                            *
      *                                                                           *
      *****************************************************************************


 Based on a SoC/HPS environment like the DE10-Nano board: MFM decoder module for 
 the DEC RL01/RL02 disk drives. 
 This module should replace  the schematics based version My_MFM_Decoder_V2_2_SoC.bdf 
         which is part of the RL01/RL02 disk reader/emulator
    
    
*/
module      my_MFM_DEcoder_SoC(
            Enable,                         // Enable
            System_clk,                     // RL System_clk, 4.1MHz
            Clk_65_6,                       // System Clock 65.6 MHz
            MFM_in,                         // MFM signal 
            Write_Gate_I,                   // data End in one sector
            Sector,                         // Start decoding (@ sector)
            //
            Clock_16bit,                    // Clock 16 Bit
            Parallel_out,                   // parallel output
            Serial_MFM,
            Serial_Clk,
            Serial_Data,                    // Serial   Output
            Write_Gate_O);
            //
input   Enable, System_clk, Clk_65_6, MFM_in, Sector, Write_Gate_I;
output  [15:0]  Parallel_out;
output  Clock_16bit, Serial_MFM, Serial_Clk;
output  Serial_Data, Write_Gate_O;
//
reg [15:0]  sync_4_1;
reg [15:0]  sync_mfm;
reg [15:0]  sync_gate;
reg [23:0]  shifter0;                       // shifter, serial->parallel
reg [31:0]  word_clock;                     // 16 Bit clock
reg [15:0]  latch_word;                     // Parallel out
//
reg clksel, FlipFlop, CLK_16bit_P;
wire clock_pulse, load_word, gatep;
//
initial
begin
    word_clock     <= 0;
    FlipFlop       <= 0;
end 
//
// 
//_-_-_-_-_-_-_-_-_-_-_-_-_   Synchronize Write_gate  -_-_-_-_-_-_-_-_-_-_-_-_
//-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_
always @ (posedge Clk_65_6)
begin
    sync_gate <= { sync_gate[14:0] , Write_Gate_I };         // shift
end
assign Write_Gate_O = sync_gate[2]; 
assign gatep = (!sync_gate[1] & sync_gate[0]);  
//        
//        
//_-_-_-_-_-_-_-_-_-_-_-_-_ Synchronize 4.1MHz clock  -_-_-_-_-_-_-_-_-_-_-_-_
//-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_
always @ (posedge Clk_65_6)
begin
    sync_4_1 <= { sync_4_1[14:0] , ~System_clk };              // shift
    clksel = (!sync_4_1[3] &  sync_4_1[2]);                    // pos edge (5/4)
end
assign clock_pulse   = clksel;
assign Serial_Clk = sync_4_1[2];
//          
//          
//-------------------- Synchronize MFM signal to FPGA clock ------------------
//----------------------------------------------------------------------------
always @ (posedge Clk_65_6)
begin
    sync_mfm <= { sync_mfm[14:0] , MFM_in };                  // shift MFM <<<
end
assign Serial_MFM = (sync_mfm[2] | sync_mfm[1]);              // Scanning MFM
//
//
//
//
//
//                                +--------------------+
//                                |----- "decode" -----|
//                                +--------------------+
//
always @ (posedge Clk_65_6)
begin
    if(Enable) begin
        if(Sector) begin                                            // sector Pulse = High
            //
            if(gatep) begin                                         // Start Write_Gate
                word_clock <= 5;                                    // 6 ist lt logic analyser am besten
            end else if( clock_pulse & Write_Gate_O ) begin         // @ Write_Gate
                if (Serial_MFM) begin
                    FlipFlop <= 1;
                end else begin
                    FlipFlop <= 0;
                end         
                shifter0 <= { shifter0[22:0] , FlipFlop};
                word_clock <= word_clock + 1; 
            end
            //
        end else begin
            FlipFlop    <= 0;
            word_clock  <= 0;                                      // Preset counter
        end
    end else begin
        FlipFlop <= 0;
        shifter0 <= 0;
    end
end
assign Clock_16bit    = word_clock[3];
assign Serial_Data    = FlipFlop;
//
//
//=============== Generate load pulse for parallel out =======================
//============================================================================
always @ (posedge Clk_65_6)
begin
    CLK_16bit_P <= Clock_16bit;
end
//assign load_word = Clock_16bit & !CLK_16bit_P;        // pos edge (V2)
assign load_word = !Clock_16bit & CLK_16bit_P;          // neg edge (V1)
//
//=========================== latch 16 bit output ============================
//============================================================================
always @ (posedge Clk_65_6)
begin
  if (load_word ) begin                                  // little endian ordering
    latch_word[15] <= shifter0[0]; 
    latch_word[14] <= shifter0[1];
    latch_word[13] <= shifter0[2];
    latch_word[12] <= shifter0[3];
    latch_word[11] <= shifter0[4];
    latch_word[10] <= shifter0[5];
    latch_word[9]  <= shifter0[6];
    latch_word[8]  <= shifter0[7];
    //
    latch_word[7]  <= shifter0[8];  
    latch_word[6]  <= shifter0[9];
    latch_word[5]  <= shifter0[10];
    latch_word[4]  <= shifter0[11];
    latch_word[3]  <= shifter0[12];
    latch_word[2]  <= shifter0[13];
    latch_word[1]  <= shifter0[14];
    latch_word[0]  <= shifter0[15];
  end
end
assign Parallel_out = latch_word;                          // 16bit out
endmodule
