/*
              *****************************************************************************
              *  Copyright (C) by Reinhard Heuberger                                      *
              *                                                                           *
              *  All rights reserved.                                                     *
              *                                                                           *
              *  Redistribution and use in source and binary forms, with or without       * 
              *  modification, are permitted provided that the following conditions       * 
              *  are met:                                                                 *
              *                                                                           * 
              *  1. Redistributions of source code must retain the above copyright        * 
              *     notice, this list of conditions and the following disclaimer.         *
              *  2. Redistributions in binary form must reproduce the above copyright     *
              *     notice, this list of conditions and the following disclaimer in the   *
              *     documentation and/or other materials provided with the distribution.  *
              *  3. Neither the name of the author nor the names of its contributors may  * 
              *     be used to endorse or promote products derived from this software     *
              *     without specific prior written permission.                            *
              *                                                                           *
              *****************************************************************************
*/
// This module generates the RL02 sector pulse. We need 2 different
// sector pulse. Pulse #1 is responsible for the Data start phase if
// this pulse is LOW. Pulse #2 is responsible for the Servo/Header
// start time if this pulse is LOW. Here is the required timing:
//
//      |<---Sector-time,625uSec = 16KHz-->|
// _____-----------_________________________-----------_______
//      |
//     >|----------------|<=== Sector-Puls #1 = 62.5 uSec
//     >|----------|<========= Pulse 48.5 uSec
//     >|-------------|<====== Pulse 53.6 uSec (57.4)
//
// 
//  Jan 2021: Sector output synchronized to system clock.
//           Sector timing #2 von 57.4 auf 53.6 geändert
//           damit beim write-one-track immer alle Daten Platz haben
//
//
module my_sector_pulse_SoC ( 
                             clk_65_6,
                             Pulse_62_5uSec_H,
                             Pulse_48_5uSec_H,
                             Pulse_48_5uSec_L,
                             Pulse_53_6uSec_H,
                             Pulse_53_6uSec_L);
//
input  clk_65_6;
output Pulse_62_5uSec_H;
output Pulse_48_5uSec_H, Pulse_48_5uSec_L; 
output Pulse_53_6uSec_H, Pulse_53_6uSec_L;
//
reg baseline, flip;
reg Pulse_62_5uSec_H, Pulse_62_5uSec_L; 
reg [31:0] count_base;
reg [31:0] count_48_5uSec;
reg [31:0] count_53_6uSec;
//
reg puls_a;
reg puls_b;
reg mypuls_a, mypuls_b;
wire pa, pb;
//
// 
// internal : Construct the baseline sector clock, 1:1 duty-cyle. 
// 65.6 MHz = 0.01524 uSec , 625 uSec = 2 mal 312.5 uSEC 
// Also 312.5 uSec / 0.01524 = 20505.24934 
// Counter läuft aber von 0, also counter-wert 20506
//
always @ ( posedge clk_65_6 ) begin
    if ( count_base <= 20506 ) begin
       count_base = count_base + 1;
    end else begin
       count_base = 0;
       flip = ~flip;
    end
end
//
always @ ( posedge clk_65_6 ) 
begin
    Pulse_62_5uSec_H  <= ~flip;
end
// 
//     Construct 2 additional sector pulse based on :
// 16bit period = 3.917uSec , sector periode = 0.625mSec
//
//
// a)65.6 MHz = 0.0152439 uSec , also:
//   (48.5uSec : 0.01524) = 3183 Counter-wert.
always @ ( posedge clk_65_6 ) begin
  if (Pulse_62_5uSec_H) begin
    if ( count_48_5uSec <= 3186 ) begin           // 3183 ( 3182,4146981 )
       count_48_5uSec = count_48_5uSec + 1;
       mypuls_a <=1;
    end else begin
       count_48_5uSec = count_48_5uSec;
       mypuls_a <=0;
    end
  end else if (!Pulse_62_5uSec_H) begin
      count_48_5uSec <= 0;
      mypuls_a <=0;
  end
end
assign pa = mypuls_a;
//
//
// b) 65.6 MHz = 0.01524 uSec , also:
//  (57_4uSec : 0.01524) = 3768 Counter-wert.
//  (53_6uSec : 0.01524) = 3517 Counter-wert.  
//  3518=bad  3530=bad 3530=bad 3510 = OK
always @ ( posedge clk_65_6 ) begin
  if (Pulse_62_5uSec_H) begin
    if ( count_53_6uSec <= 3510 ) begin            // 3768 ( 3765,4406024 )
       count_53_6uSec = count_53_6uSec + 1;
       mypuls_b <=1;
    end else begin
       count_53_6uSec = count_53_6uSec;
       mypuls_b <=0;
    end
  end else if (!Pulse_62_5uSec_H) begin
      count_53_6uSec <= 0;
      mypuls_b <=0;
  end
end
assign pb = mypuls_b; 
//
//
// synchronizee to system clock
always @ ( posedge clk_65_6 ) begin
    puls_a <= pa;
end
assign Pulse_48_5uSec_H =  puls_a; 
assign Pulse_48_5uSec_L = ~puls_a;
//
always @ ( posedge clk_65_6 ) begin
    puls_b <= pb;
end
assign Pulse_53_6uSec_H =  puls_b;
assign Pulse_53_6uSec_L = ~puls_b;
//
//
endmodule