/*
      *****************************************************************************
      *  Copyright (C) by Reinhard Heuberger , www.pdp11gy.com                    *
      *                                                                           *
      *  All rights reserved.                                                     *
      *                                                                           *
      *  Redistribution and use in source and binary forms, with or without       *
      *  modification, are permitted provided that the following conditions       *
      *  are met:                                                                 *
      *                                                                           *
      *  1. Redistributions of source code must retain the above copyright        *
      *     notice, this list of conditions and the following disclaimer.         *
      *  2. Redistributions in binary form must reproduce the above copyright     *
      *     notice, this list of conditions and the following disclaimer in the   *
      *     documentation and/or other materials provided with the distribution.  *
      *  3. Neither the name of the author nor the names of its contributors may  *
      *     be used to endorse or promote products derived from this software     *
      *     without specific prior written permission.                            *
      *                                                                           *
      *****************************************************************************
        
 Einen kompletten track schreiben, auf emulierten und realen RL01/RL02 disk drive.
 Der zu schreibende Track wird zuerst mal eingelesen und die Sectoren müssen von 
 0 bis 39 aufsteigend sortiert werden. Das wird in der C-Routine "reorder_track" 
 erledigt. Nun wird Write/Port2 auf das DPR verhindert und der geordnete Track ins 
 DPR kopiert. Die benötigte header/servo struktur der zu beschreibenden Disk steht 
 am Ausgang des RL_MFM_decoder zur Verfügung. Die zu schreibenden Daten befinden
 sich im DPR. Nun wird der sector 0 gesucht und das Write Gate Signal generiert.
 Die Daten werden vom MFM ENcoder aus zur Disk gesendet. Würde man nur emulierte
 RL drives benützen, wäre es total einfach mit einen memcpy Befehl machbar, aber
 ich habe es mir nun mal in den Kopf gesetzt, auch auf eine Reale RL zu schreiben.
 Bin nur mal gespannt, ob man bei einer realen RL einen kompletten Track auf einmal
 schreiben kann.

 This module consists of 2 parts: 
 1) 2 MFM ENCODER with parallel to serial converter. Details about time and phase
    shifting can be found in the module "my_MFM_ENcoder_SoC.v"
 2) Write-sequencer: generate all signals to write one track on the disk. It makes  
    no difference whether the RL01/02-disk is emulated or real. 
//
//                    Version 2.8E / JANUAR 2021
//           part of the RL01/RL02 disk reader/emulator Version V2.8x
*/ 
module write_one_track(
        Enable,
        from_ram_data_16_in,        // Data, 16 Bit parallel in
        from_dsk_data_16_in,        // Data, 16 Bit parallel in
        sector,                     // Start if sector High.
        clk_65_6,                   // Input Clock  65.6 MHZ
        first_MFM,
        //
        CLK_16bit,                  // Out-clock, 16Bit
        address_out, 
        serial_OUT_dsk,             // Data, Serial out
        serial_OUT_ram,             // Data, Serial out
        MFM,                        // MFM out
        TST,
        go,                         // 
        W_gate,                     // Write Gate
        sysclk);
//
input Enable;
input [15:0] from_ram_data_16_in;
input [15:0] from_dsk_data_16_in;      
input sector, clk_65_6, first_MFM; 
//
output CLK_16bit; 
output [12:0] address_out; 
output serial_OUT_dsk, serial_OUT_ram;
output MFM, TST, go, W_gate, sysclk;
//
reg [31:0]  divider;         // counter,= clock generator.
reg [31:0]  count;           // Pulse counter
reg [15:0]  shifter_dsk;     // 16Bit parallel in SR from disk
reg [15:0]  shifter_ram;     // 16Bit parallel in SR from ram
reg [15:0]  history_dsk;     // jetzt, alt, uralt SR.
reg [15:0]  history_ram;     // jetzt, alt, uralt SR.
reg [15:0]  in;              // Parallel input
//
reg [31:0]  address;
reg [31:0]  base_adr;
//
reg [15:0]  write_SR;        // load-puls generator write part.
reg [2:0]   rclksync_16_4;   // like 2 D-FF register, 
reg clk_4_1_p;               // Synchronized 4.1MHz Clock
reg MFM, ph0, ph1, ph3;
reg go, track, W_gate ; 
reg PR,  PR1flag, sector_p;
//
wire load_L;
wire clockw;
wire load2, load3, load4, CLK_8_2, CLK_4_1; 
wire phase_0, phase_1, phase_3, addrpuls; 
wire shifter_dsk_out, shifter_ram_out;
//
initial
begin
    divider         <= 0;
    shifter_dsk     <= 0;
    history_dsk     <= 0;
    W_gate          <= 0;
    go              <= 0;    // 1, wenn ein track geschrieben wird
    track           <= 0;    // 1, wenn ein track geschrieben wurde.
    base_adr        <= 0;
    PR1flag         <= 1;
end
//
//
//========================== Counter/Divider =================================
//============================================================================
always @ (posedge clk_65_6)
begin
    if(sector & Enable & first_MFM ) begin
        divider <= divider - 1;
    end else if(!first_MFM & !sector) begin
        divider <= 0;
    end
end
assign CLK_8_2   =  divider[2];             // Output: 8.2Mhz
assign CLK_4_1   =  divider[3];             // Output: 4.1MHz
assign clockw    =  divider[7];             // 16Bit Clock
assign sysclk    = ~divider[3];
//
//============== MFM phase and pulse-width generator =========================
//============================================================================
always @ (posedge clk_65_6)
begin
 ph0 <= 0;
 ph1 <= ( CLK_4_1 &&  CLK_8_2);
 ph3 <= (!CLK_4_1 &&  CLK_8_2);
end
assign phase_0 = ph0;
assign phase_1 = ph1;
assign phase_3 = ph3;
//
//
//======================== Generate load pulses ==============================
//============================================================================
always @ (posedge clk_65_6)
begin
    write_SR = {write_SR[14:0], clockw };   // shift
end
assign CLK_16bit = clockw;                          // Output
assign load_L = ~((write_SR[1] ^ clockw) & clockw); // Encoder 
assign load2 = (write_SR[1] & !write_SR[0]);        // Writer: neg edge
assign load3 = (write_SR[2] & !write_SR[1]);        // Writer: neg edge
assign load4 = (write_SR[3] & !write_SR[2]);        // Writer: neg edge
//
//
//========================= Mofify 4.1MHz clock  =============================
//============================================================================
always @ (posedge clk_65_6)
begin
 clk_4_1_p <= CLK_4_1;
end
//
//
//=============== Shiftregister with synchronous parallel load ===============
//============================================================================
  always @ (posedge clk_65_6)                       
  begin
   if(!sector & !first_MFM )
     shifter_dsk = 0;
   else if ( !clk_4_1_p & CLK_4_1 )
     if (!load_L ) 
       shifter_dsk <= from_dsk_data_16_in;
     else if (load_L) 
       shifter_dsk <= shifter_dsk >> 1;
  end
 //
assign  shifter_dsk_out  = shifter_dsk[0];
//-----------------------------------------------------
 always @ (posedge clk_65_6)                       
  begin
   if(!sector  & !first_MFM  )
     shifter_ram = 0;
   else if ( !clk_4_1_p & CLK_4_1 )
     if (!load_L ) 
       shifter_ram <= from_ram_data_16_in;
     else if (load_L) 
       shifter_ram <= shifter_ram >> 1;
  end
 //
assign  shifter_ram_out  = shifter_ram[0];
//
//
//========================= history_dsk shiftregister  ===========================
//============================================================================
always @ (posedge clk_65_6)
begin
 if ( !clk_4_1_p & CLK_4_1 )
 begin
  //
  history_dsk = {history_dsk[14:0], shifter_dsk_out};   // shift
  history_ram = {history_ram[14:0], shifter_ram_out};   // shift
  //
 end
end
assign  serial_OUT_dsk  = history_dsk[7];          // Serial out @ byte boundary 
assign  serial_OUT_ram  = history_ram[7];          // Serial out @ byte boundary        
//
//
//============================ MFM - ENcoder =================================
//============================================================================
// MFM spezifisch: Von 1 ->  einer(!) 0 mit folgender 1 :
// 1 auf 0  wird immer Phase_0, aber um die darauf folgende
// richtig zu encoden, muss eine Phase_3 gesendet werden. 
always @(posedge clk_65_6)
begin
  if (sector) begin
    case (history_ram[9:7])                             //!!!!!!!!!!!!!!!!!!!!!!
      //   |
      3'b000: MFM <= phase_1;         // 00->0 
      3'b001: MFM <= phase_3;         // 00->1 
      3'b010: MFM <= phase_0;         // 01->0 
      3'b011: MFM <= phase_3;         // 01->1 
      3'b100: MFM <= phase_1;         // 10->0 
      3'b101: MFM <= phase_3;         // 10->1 !
      3'b110: MFM <= phase_0;         // 11->0 
      3'b111: MFM <= phase_3;         // 11->1  
      //
    default: MFM <= 0;    
    endcase
  end
  else begin
    MFM <= 0;
  end
end
//
//
//
//
//                                +***************************+
//                                |----- write-sequencer -----|
//                                +***************************+
//
//
always @ (posedge clk_65_6)
begin
    sector_p <= sector;
end
assign addrpuls = !sector_p & sector;                            // puls @ sector_H
//
//
//-----------------------------------------------------------------------------------------------------------+
always @ (posedge clk_65_6)
begin
    if(Enable) begin
        if( addrpuls ) begin
            if (go) begin
                base_adr  <= base_adr +144;
                address   <= base_adr; 
            end else if (!go) begin 
                base_adr  <= 0;
                address   <= 0;
            end
        end
        //
        if(sector) begin
//-----------------------------------------------------------------------------------------------------------+
            if (load2) begin                                       // Load2 pulse ?
                in <= from_dsk_data_16_in;                         // Get 16 bit data
                count <= count +1;                                 // increment counter
//-----------------------------------------------------------------------------------------------------------+
            end else if (load3) begin                              // Load3 pulse ?
                if(in == 16'h8000 & !PR1flag) begin                // PR1 or PR2 Value , 0x8000
                    PR       <= 1;
                    PR1flag  <= 1;
                    count    <= 2;                                  //
                    address  <= address + 2;                        // Adress to PR1
                end else if (PR1flag & (count == 3)) begin          // if PR1, next word contains sec.nr.                        
                    if((in & 16'h003F) == 16'h0000) begin           // ********* Sector 0 ? **********
                        if(track) begin                             // einmal durch !
                            go             <= 0;
                        end else if(!track) begin
                            track          <= 1;                    //Set: einmal durch
                            go             <= 1;
                        end
                    end
                end else if(in == 16'h8000 & PR1flag & go) begin    // if PR2
                    PR       <= 0;                                  //
                end else if((count == 5) & go) begin                // if PD1 !!! mit go!!
                    W_gate  <= 1;                                   // *****Set Write Gate*******
                end
//-----------------------------------------------------------------------------------------------------------+
             end else if (load4) begin                              // Load4 pulse ? 
                if(count < 2) begin
                    address <= base_adr;           
                end else if((count >= 2) & ( count <= 140) ) begin
                    address <= address + 1;             
                end else if( count > 140) begin                    // >= 140
                    W_gate      <= 0;
                    count       <= 0;
                end
            end             
//-----------------------------------------------------------------------------------------------------------+
        end else if(!sector ) begin                                 // neg. sector pulse 
            PR          <= 0;
            PR1flag     <= 0;
            W_gate      <= 0;
            count       <= -10;                                     // -10(V1.8)
        end
    end else if(!Enable) begin                                      // Enable ?
        go              <= 0;
        track           <= 0;
        address         <= 0;
        base_adr        <= 0;
        W_gate          <= 0;
        count           <= -10; 
    end         
end
assign address_out = address[12:0];
assign TST = PR;
//
//
endmodule
